from . import base

import numpy as np


class StandardScaleRegressor(base.Regressor):
    """Meta-regressor that rescales the target variable.

    A running mean and standard deviation are maintained and used to modify the target variable
    before feeding it to the underlying regressor. The output of the regressor is unscaled before
    being given to the user.

    """

    def __init__(self, regressor=None, eps=None):
        self.regressor = regressor
        self.count = 0
        self.mean = 0
        self.sos = 0
        self.eps = eps or np.finfo(float).eps

    def _rescale(self, y):
        return (y - self.mean) / (self.eps + self.sos / self.count) ** 0.5

    def _unscale(self, y):
        return y * (self.eps + self.sos / self.count) ** 0.5 + self.mean

    def fit_one(self, x, y):

        self.count += 1
        mean = self.mean
        self.mean += (y - mean) / self.count
        self.sos += (y - mean) * (y - self.mean)

        return self._unscale(self.regressor.fit_one(x, self._rescale(y)))

    def predict_one(self, x):
        return self._unscale(self.regressor.predict_one(x))



"""

- https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/lm/sgd_fast.pyx
"""
import abc
import math


__all__ = ['SquaredLoss', 'LogLoss']


class Loss(abc.ABC):

    @abc.abstractmethod
    def __call__(self, y_true, y_pred) -> float:
        pass

    @abc.abstractmethod
    def gradient(self, y_true, y_pred) -> dict:
        pass


class SquaredLoss(Loss):

    def __call__(self, y_true, y_pred):
        return (y_pred - y_true) ** 2

    def gradient(self, y_true, y_pred):
        return 2 * (y_pred - y_true)


class LogLoss(Loss):

    @staticmethod
    def clip_proba(p):
        return max(min(p, 1 - 1e-15), 1e-15)

    def __call__(self, y_true, y_pred):
        y_pred = self.clip_proba(y_pred)
        return -(y_true * math.log(y_pred) + (1 - y_true) * math.log(1 - y_pred))

    def gradient(self, y_true, y_pred):
        return self.clip_proba(y_pred) - y_true



import abc

import numpy as np
try:
    import pandas as pd
    PANDAS_INSTALLED = True
except ImportError:
    PANDAS_INSTALLED = False
from sklearn import base
from sklearn import utils

from . import stream


STREAM_METHODS = {
    np.ndarray: stream.iter_numpy
}

if PANDAS_INSTALLED:
    STREAM_METHODS[pd.DataFrame] = stream.iter_pandas


class Regressor(base.BaseEstimator, base.RegressorMixin):

    @abc.abstractmethod
    def fit_one(x, y):
        """Fits to set of features `x` and a real-valued target `y` and returns the estimated
        target value of `x` before seeing `y`.
        """
        pass

    def fit(self, X, y=None):
        X = utils.check_array(X)
        for x, yi in STREAM_METHODS[type(X)](X, y):
            self.fit_one(x, yi)
        return self

    @abc.abstractmethod
    def predict_one(x: dict) -> float:
        pass

    def predict(self, X):
        X = utils.check_array(X)
        y_pred = np.empty(len(X), dtype='float')
        for i, (x, _) in enumerate(stream.iter_numpy(X, shuffle=False)):
            y_pred[i] = self.predict_one(x)
        return y_pred


class BinaryClassifier(base.BaseEstimator, base.ClassifierMixin):

    @abc.abstractmethod
    def fit_one(x, y):
        """Fits """
        pass

    def fit(self, X, y=None):
        X = utils.check_array(X)
        for x, yi in STREAM_METHODS[type(X)](X, y):
            self.fit_one(x, yi)
        return self

    @abc.abstractmethod
    def predict_proba_one(x: dict) -> float:
        pass

    @abc.abstractmethod
    def predict_one(x: dict) -> bool:
        pass

    def predict_proba(self, X):
        X = utils.check_array(X)
        y_pred = np.empty(len(X))
        for i, (x, _) in enumerate(stream.iter_numpy(X, shuffle=False)):
            y_pred[i] = self.predict_proba_one(x)
        return y_pred

    def predict(self, X):
        X = utils.check_array(X)
        return np.round(self.predict_proba(X)).astype(bool)


class MultiClassifier(base.BaseEstimator, base.ClassifierMixin):

    @abc.abstractmethod
    def fit_one(x, y):
        """Fits """
        pass

    def fit(self, X, y=None):
        X = utils.check_array(X)
        for x, yi in STREAM_METHODS[type(X)](X, y):
            self.fit_one(x, yi)
        return self

    @abc.abstractmethod
    def predict_proba_one(x: dict) -> dict:
        pass

    @abc.abstractmethod
    def predict_one(x: dict) -> str:
        pass

    def predict_proba(self, X):
        X = utils.check_array(X)
        y_pred = [None] * len(X)
        for i, (x, y) in enumerate(stream.iter_numpy(X, shuffle=False)):
            y_pred[i] = list(self.fit_one(x, y).values())
        return np.array(y_pred)

    def predict(self, X):
        X = utils.check_array(X)
        y_pred = [None] * len(X)
        for i, (x, _) in enumerate(stream.iter_numpy(X, shuffle=False)):
            y_pred[i] = self.predict_one(x)
        return np.array(y_pred)


class Transformer(base.BaseEstimator, base.TransformerMixin):

    @abc.abstractmethod
    def fit_one(x, y):
        """Fits """
        pass

    def fit(self, X, y=None):
        X = utils.check_array(X)
        for x, yi in STREAM_METHODS[type(X)](X, y):
            self.fit_one(x, yi)
        return self

    @abc.abstractmethod
    def transform_one(x: dict) -> dict:
        pass

    def transform(self, X):
        X = utils.check_array(X)
        y_trans = [None] * len(X)
        for i, (x, _) in enumerate(stream.iter_numpy(X)):
            y_trans[i] = list(self.transform_one(x).values())
        return np.array(y_trans)


class Clusterer(base.BaseEstimator, base.ClusterMixin):

    @abc.abstractmethod
    def fit_one(x, y):
        """Fits """
        pass

    def fit(self, X, y=None):
        X = utils.check_array(X)
        for x, yi in STREAM_METHODS[type(X)](X, y):
            self.fit_one(x, yi)
        return self

    @abc.abstractmethod
    def predict_one(x: dict) -> dict:
        pass

    def predict(self, X):
        X = utils.check_array(X)
        y_pred = np.empty(len(X), dtype='int32')
        for i, (x, _) in enumerate(stream.iter_numpy(X, shuffle=False)):
            y_pred[i] = self.predict_one(x)
        return y_pred



import collections
import functools
import math
import re

from sklearn import feature_extraction

from .. import base


__all__ = ['CountVectorizer', 'TFIDFVectorizer']


def compose(*functions):
    """Return a callable that chains multiple functions.

    Example
    -------

    >>> f = lambda x: x + 1
    >>> g = lambda x: x * 2
    >>> h = lambda x: -x
    >>> compose(f, g, h)(10)
    -22

    """

    def compose2(f, g):
        return lambda x: f(g(x))

    return functools.reduce(compose2, reversed(functions), lambda x: x)


class VectorizerMixin:

    def __init__(self, on: str, strip_accents=True, lowercase=True, preprocessor=None,
                 tokenizer=None):
        self.on = on
        self.strip_accents = strip_accents
        self.lowercase = lowercase
        self.preprocessor = preprocessor or self.build_preprocessor()
        self.tokenizer = tokenizer or self.build_tokenizer()

    def build_preprocessor(self):
        """Return a function to preprocess the text before tokenization."""
        steps = []

        if self.strip_accents:
            steps.append(feature_extraction.text.strip_accents_unicode)

        if self.lowercase:
            steps.append(str.lower)

        return compose(*steps) if steps else lambda x: x

    def build_tokenizer(self):
        """Return a function that splits a string into a sequence of tokens."""
        token_pattern = re.compile(r'(?u)\b\w\w+\b')
        return lambda text: token_pattern.findall(text)


class CountVectorizer(base.Transformer, VectorizerMixin):
    """
    Example
    -------

        >>> import creme
        >>> corpus = [
        ...     'This is the first document.',
        ...     'This document is the second document.',
        ...     'And this is the third one.',
        ...     'Is this the first document?',
        ... ]
        >>> vectorizer = creme.feature_extraction.CountVectorizer(on='sentence')
        >>> for sentence in corpus:
        ...     print(vectorizer.fit_one({'sentence': sentence}))
        Counter({'this': 1, 'is': 1, 'the': 1, 'first': 1, 'document': 1})
        Counter({'document': 2, 'this': 1, 'is': 1, 'the': 1, 'second': 1})
        Counter({'and': 1, 'this': 1, 'is': 1, 'the': 1, 'third': 1, 'one': 1})
        Counter({'is': 1, 'this': 1, 'the': 1, 'first': 1, 'document': 1})

    """

    def fit_one(self, x, y=None):
        return self.transform_one(x)

    def transform_one(self, x):
        return collections.Counter(self.tokenizer(self.preprocessor(x[self.on])))


class TFIDFVectorizer(base.Transformer, VectorizerMixin):

    def __init__(self, normalize=True, **kwargs):
        super().__init__(**kwargs)
        self.normalize = normalize
        self.counter = CountVectorizer(**kwargs)
        self.dfs = collections.defaultdict(int)
        self.n = 0

    def compute_tfidfs(self, term_counts):
        n_terms = sum(term_counts.values())

        tfidfs = {}

        for term, count in term_counts.items():
            tf = count / n_terms
            idf = math.log(((1 + self.n) / (1 + self.dfs[term]))) + 1
            tfidfs[term] = tf * idf

        if self.normalize:
            norm = math.sqrt(sum(tfidf ** 2 for tfidf in tfidfs.values()))
            return {term: tfidf / norm for term, tfidf in tfidfs.items()}
        return tfidfs

    def fit_one(self, x, y=None):

        # Compute the term counts
        term_counts = self.counter.fit_one(x)

        # Increment the document counter
        self.n += 1

        for term in term_counts:
            self.dfs[term] += 1

        return self.compute_tfidfs(term_counts)

    def transform_one(self, x):
        term_counts = self.counter.transform_one(x)
        return self.compute_tfidfs(term_counts)



import abc


__all__ = ['ConstantLR', 'LinearDecreaseLR']


class LRScheduler(abc.ABC):

    def __init__(self, init_lr):
        self.init_lr = init_lr

    @abc.abstractmethod
    def get(self, t: int) -> float:
        pass


class ConstantLR(LRScheduler):

    def get(self, t):
        return self.init_lr


class LinearDecreaseLR(LRScheduler):
    """Implements the decreasing LR schedule proposed by Léon Bottou.

    See somewhere around line 64 of https://leon.bottou.org/git/?p=sgd.git;a=blob;f=README.txt for
    some details.

    """

    def __init__(self, init_lr, strengh):
        super().__init__(init_lr)
        self.strengh = strengh

    def get(self, t):
        return self.init_lr / (1 + t) ** self.strengh



"""
## Introduction

`creme` is a library for in**creme**ntal learning. Incremental learning is a machine learning
regime where the observations are made available one by one. It is also known as online learning,
iterative learning, or sequential learning. This is in contrast to batch learning where all the
data is processed at once. Incremental learning is desirable when the data is too big to fit in
memory, or simply when it isn't available all at once. `creme`'s API is heavily inspired from that
of [scikit-learn](https://scikit-learn.org/stable/), enough so that users who are familiar with
scikit-learn should feel right at home.

Most machine learning algorithms (be it supervised or unsupervised) assume a batch regime. However
some of these algorithms have online variants. For example
[stochastic gradient descent](https://www.wikiwand.com/en/Stochastic_gradient_descent) is the
online version of [gradient descent](https://www.wikiwand.com/en/Gradient_descent) whilst
[incremental k-means clustering](http://www.cs.princeton.edu/courses/archive/fall08/cos436/Duda/C/sk_means.htm)
is the online adaptation of [k-means clustering](https://www.wikiwand.com/en/K-means_clustering).
In general online algorithms perform slightly worse than their batch counterparts, although the gap
is usually very small. However, online learning algorithms only consume a tiny amount of RAM, which
thus makes them scalable and ideal candidates for commodity hardware and embedded systems.

The objective of `creme` is to provide a nice interface for putting an incremental learning
pipeline in place; a bit like what scikit-learn does for batch learning. Of course there are other
open-source solutions available, but they are somewhat specialized towards certain tasks and can
require a steep learning curve. Moreover some of these solutions aren't "truly online" as they
mostly assume the data is contained in a file. With `creme` it is possible to learn from a stream
in it's largest sense, be it a database query or a Kafka instance.

## API

Just like scikit-learn, each of `creme`'s estimators provide a similar API. Every estimator has a
`fit_one(x, y)` method which will fit the estimator with a given set of features `x` and a
target `y`. In addition, estimators have a `predict_one(x)` or a `transform_one(x)` method,
depending on their type. Classifiers also implement a `predict_proba_one(x)` method. Each call to
`fit_one` will also return the predicted value for the current `x`, which makes it possible to
monitor the progress of the estimator online. Although `creme`'s purpose is incremental learning,
each of it's estimators also implements the scikit-learn's `fit/predict/transform` API, which makes
it possible to reuse scikit-learn's toolbox. This is mostly intended to help comparing batch
algorithms to their online versions.

Rows in `creme` are represented by `dict`s that map feature names to values. The main advantage of
using a `dict` over a `numpy` array is that features can be accessed by name rather by position.
Moreover `dict`s can store values of different types, whereas all the values in a `numpy` array
have to be of a single type. What's more `dict`s are sparse by default; indeed a feature with a
`None` value can simply be omitted. Finally Python has some nice tools for working with `dict`s,
such as `collections.defaultdict`.

The `creme` library is organized in modules, following the fashion in which scikit-learn is
organized. There is thus a `preprocessing` module as well as a `feature_extraction` module, amongst
others. Furthermore an effort is made to keep the naming of `creme`'s classes and parameters
consistent with scikit-learn.

For more information please check out the [documentation](https://creme.github.io).

## Important links

- Documentation
- Example notebooks
- Issue tracker
- Releases

## Installation

```sh
pip install creme
```

## Quick example

In the following snippet we'll be fitting an online logistic regression. The weights of the model
will be optimized with the [AdaGrad](http://akyrillidis.github.io/notes/AdaGrad) algorithm. We'll
scale the data so that each variable has a mean of 0 and a standard deviation of 1. The standard
scaling and the logistic regression are combined using a pipeline. We'll be using the
`stream.iter_sklearn_dataset` function for streaming over the
[Wisconsin breast cancer dataset](http://archive.ics.uci.edu/ml/datasets/breast+cancer+wisconsin+%28diagnostic%29).
We'll measure the ROC AUC using [progressive validation](http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.153.3925&rep=rep1&type=pdf).

```python
>>> from creme import linear_model
>>> from creme import model_selection
>>> from creme import optim
>>> from creme import pipeline
>>> from creme import preprocessing
>>> from creme import stream
>>> from sklearn import datasets
>>> from sklearn import metrics

>>> X_y = stream.iter_sklearn_dataset(
...     load_dataset=datasets.load_breast_cancer,
...     shuffle=True,
...     random_state=42
... )
>>> optimizer = optim.AdaGrad()
>>> model = pipeline.Pipeline([
...     ('scale', preprocessing.StandardScaler()),
...     ('learn', linear_model.LogisticRegression(optimizer))
... ])
>>> metric = metrics.roc_auc_score

>>> model_selection.online_score(X_y, model, metric)
0.992977...

```

"""
from .__version__ import __version__

__all__ = [
    'cluster',
    'compose',
    'ensemble',
    'feature_extraction',
    'linear_model',
    'losses',
    'model_selection',
    'naive_bayes',
    'optim',
    'pipeline',
    'preprocessing',
    'reco',
    'stats',
    'stream',
    'tree'
]



import collections

from . import base


__all__ = ['RMSProp']


class RMSProp(base.Optimizer):
    """
    Example
    -------

        #!python
        >>> import creme
        >>> from sklearn import datasets
        >>> from sklearn import metrics

        >>> X_y = creme.stream.iter_sklearn_dataset(
        ...     load_dataset=datasets.load_breast_cancer,
        ...     shuffle=True,
        ...     random_state=42
        ... )
        >>> optimiser = creme.optim.RMSProp()
        >>> model = creme.pipeline.Pipeline([
        ...     ('scale', creme.preprocessing.StandardScaler()),
        ...     ('learn', creme.linear_model.LogisticRegression(optimiser))
        ... ])
        >>> metric = metrics.roc_auc_score

        >>> creme.model_selection.online_score(X_y, model, metric)
        0.991167...

    """

    def __init__(self, lr=0.1, rho=0.9, eps=1e-8):
        super().__init__(lr)
        self.rho = rho
        self.eps = eps
        self.g2 = collections.defaultdict(lambda: 0.)

    def update_weights_with_gradient(self, w, g):

        for i, gi in g.items():
            self.g2[i] = self.rho * self.g2[i] + (1 - self.rho) * gi ** 2
            w[i] -= self.learning_rate / (self.g2[i] + self.eps) ** 0.5 * gi

        return w



import collections
import copy

from sklearn import utils

from .. import base


__all__ = ['BaggingClassifier']


class BaggingClassifier(base.BinaryClassifier):
    """Bagging for classification.

    For each incoming observation, each model's `fit_one` method is called `k` times where `k`
    is sampled from a Poisson distribution of parameter 1. `k` thus has a 36% chance of being equal
    to 0, a 36% chance of being equal to 1, an 18% chance of being equal to 2, a 6% chance of being
    equal to 3, a 1% chance of being equal to 4, etc. You can do `scipy.stats.poisson(1).pmf(k)`
    for more detailed values.

    In the following example three logistic regressions are bagged together. The performance is
    slightly better than when one using a single logistic regression.

        >>> import creme.ensemble
        >>> import creme.linear_model
        >>> import creme.model_selection
        >>> import creme.optim
        >>> import creme.pipeline
        >>> import creme.preprocessing
        >>> import creme.stream
        >>> from sklearn import datasets
        >>> from sklearn import metrics

        >>> X_y = creme.stream.iter_sklearn_dataset(
        ...     load_dataset=datasets.load_breast_cancer,
        ...     shuffle=True,
        ...     random_state=42
        ... )
        >>> optimiser = creme.optim.VanillaSGD()
        >>> model = creme.pipeline.Pipeline([
        ...     ('scale', creme.preprocessing.StandardScaler()),
        ...     ('learn', creme.linear_model.LogisticRegression(optimiser))
        ... ])
        >>> model = creme.ensemble.BaggingClassifier(model, n_estimators=3)
        >>> metric = metrics.roc_auc_score

        >>> creme.model_selection.online_score(X_y, model, metric)
        0.991563...

    - [Online Bagging and Boosting](https://ti.arc.nasa.gov/m/profile/oza/files/ozru01a.pdf)

    """

    def __init__(self, base_estimator=None, n_estimators=10, random_state=42):
        self.base_estimator = base_estimator
        '''The estimator to bag, should implement `creme.base.Classifier`.'''
        self.n_estimators = n_estimators
        self.estimators = [copy.deepcopy(base_estimator) for _ in range(n_estimators)]
        self.rng = utils.check_random_state(random_state)

    def fit_one(self, x, y):

        y_pred = self.predict_proba_one(x)

        for estimator in self.estimators:
            for _ in range(self.rng.poisson(1)):
                estimator.fit_one(x, y)

        return y_pred

    def predict_one(self, x):
        votes = collections.Counter((estimator.predict_one(x) for estimator in self.estimators))
        return max(votes, key=votes.get)

    def predict_proba_one(self, x):
        return sum(estimator.predict_proba_one(x) for estimator in self.estimators) / len(self.estimators)



from sklearn import utils
import itertools


__all__ = ['iter_numpy', 'iter_pandas', 'iter_sklearn_dataset']


def iter_numpy(X, y=None, feature_names=None, shuffle=False, random_state=None):
    feature_names = list(range(X.shape[1])) if feature_names is None else feature_names
    rng = utils.check_random_state(random_state)

    if shuffle:
        order = rng.permutation(len(X))
        X, y = X[order], y if y is None else y[order]

    for x, yi in itertools.zip_longest(X, [] if y is None else y):
        yield {i: xi for i, xi in zip(feature_names, x)}, yi


def iter_sklearn_dataset(load_dataset, **kwargs):
    dataset = load_dataset()
    kwargs['X'] = dataset.data
    kwargs['y'] = dataset.target
    kwargs['feature_names'] = dataset.feature_names

    for x, yi in iter_numpy(**kwargs):
        yield x, yi


def iter_pandas(X, y=None, **kwargs):
    kwargs['feature_names'] = X.columns
    for x, yi in iter_numpy(X, y, **kwargs):
        yield x, yi



import copy

from .. import base


__all__ = ['OneVsRestClassifier']


class OneVsRestClassifier(base.MultiClassifier):
    """
    Examples
    --------

        #!python
        >>> import creme.linear_model
        >>> import creme.model_selection
        >>> import creme.multiclass
        >>> import creme.optim
        >>> import creme.pipeline
        >>> import creme.preprocessing
        >>> import creme.stream
        >>> from sklearn import datasets
        >>> from sklearn import metrics

        >>> X_y = creme.stream.iter_sklearn_dataset(
        ...     load_dataset=datasets.load_iris,
        ...     shuffle=True,
        ...     random_state=42
        ... )
        >>> optimr = creme.optim.RMSProp()
        >>> model = creme.pipeline.Pipeline([
        ...     ('scale', creme.preprocessing.StandardScaler()),
        ...     ('learn', creme.multiclass.OneVsRestClassifier(
        ...         base_estimator=creme.linear_model.LogisticRegression(optimr))
        ...     )
        ... ])
        >>> metric = metrics.accuracy_score

        >>> creme.model_selection.online_score(X_y, model, metric)
        0.813333...

    """

    def __init__(self, base_estimator):
        self.base_estimator = base_estimator
        self.models = {}

    def _normalize_preds(self, y):
        ys = sum(y.values())
        return {c: p / ys for c, p in y.items()}

    def fit_one(self, x, y):
        if y not in self.models:
            self.models[y] = copy.deepcopy(self.base_estimator)
        y_pred = {c: model.fit_one(x, y == c) for c, model in self.models.items()}
        return self._normalize_preds(y_pred)

    def predict_proba_one(self, x):
        y_pred = {c: model.predict_proba_one(x) for c, model in self.models.items()}
        return self._normalize_preds(y_pred)

    def predict_one(self, x):
        y_pred = self.predict_proba_one(x)
        return max(y_pred, key=y_pred.get)



from . import base


__all__ = ['VanillaSGD']


class VanillaSGD(base.Optimizer):
    """
    Example
    -------

        #!python
        >>> import creme
        >>> from sklearn import datasets
        >>> from sklearn import metrics

        >>> X_y = creme.stream.iter_sklearn_dataset(
        ...     load_dataset=datasets.load_breast_cancer,
        ...     shuffle=True,
        ...     random_state=42
        ... )
        >>> optimiser = creme.optim.VanillaSGD()
        >>> model = creme.pipeline.Pipeline([
        ...     ('scale', creme.preprocessing.StandardScaler()),
        ...     ('learn', creme.linear_model.LogisticRegression(optimiser))
        ... ])
        >>> metric = metrics.roc_auc_score

        >>> creme.model_selection.online_score(X_y, model, metric)
        0.990625...

    """

    def __init__(self, lr=0.1):
        super().__init__(lr)

    def update_weights_with_gradient(self, w, g):

        for i, gi in g.items():
            w[i] -= self.learning_rate * gi

        return w



"""Module for computing running statistics such as the mean and the variance.

Each running statistic has an `update` method and a `get` method.
"""
import abc


class RunningStatistic(abc.ABC):

    @property
    @abc.abstractmethod
    def name(self):
        pass

    @abc.abstractmethod
    def update(self):
        pass

    @abc.abstractmethod
    def get(self) -> float:
        pass


class Count(RunningStatistic):

    def __init__(self):
        super().__init__()
        self.n = 0

    @property
    def name(self):
        return 'count'

    def update(self, x=None):
        self.n += 1
        return self

    def get(self):
        return self.n


class Mean(RunningStatistic):

    def __init__(self):
        super().__init__()
        self.count = Count()
        self.mu = 0

    @property
    def name(self):
        return 'mean'

    def update(self, x):
        self.mu += (x - self.mu) / self.count.update().get()
        return self

    def get(self):
        return self.mu


class SmoothMean(Mean):
    """Computes the running mean using additive smoothing.

    - https://www.wikiwand.com/en/Additive_smoothing
    """

    def __init__(self, prior, prior_weight):
        super().__init__()
        self.prior = prior
        self.prior_weight = prior_weight

    @property
    def name(self):
        return 'smooth_mean'

    def get(self):
        numerator = self.mu * self.count.get() + self.prior * self.prior_weight
        denominator = self.count.get() + self.prior_weight
        return numerator / denominator


class Variance(RunningStatistic):

    def __init__(self):
        super().__init__()
        self.mean = Mean()
        self.sos = 0

    @property
    def name(self):
        return 'variance'

    def update(self, x):
        old_mean = self.mean.get()
        new_mean = self.mean.update(x).get()
        self.sos += (x - old_mean) * (x - new_mean)
        return self

    def get(self):
        return self.sos / self.mean.count.n if self.sos else 0



from . import function


__all__ = ['OneHotEncoder']


class OneHotEncoder(function.FunctionTransformer):
    """
    Example
    -------

        #!python
        >>> import pprint as pp
        >>> import string
        >>> import creme
        >>> import numpy as np

        >>> rng = np.random.RandomState(42)
        >>> alphabet = list(string.ascii_lowercase)
        >>> X = [{'letter': letter} for letter in rng.choice(alphabet, size=10)]

        >>> one_hot = creme.preprocessing.OneHotEncoder()
        >>> pp.pprint([one_hot.fit_one(x) for x in X])
        [{'letter_g': True},
         {'letter_t': True},
         {'letter_o': True},
         {'letter_k': True},
         {'letter_h': True},
         {'letter_u': True},
         {'letter_g': True},
         {'letter_z': True},
         {'letter_s': True},
         {'letter_w': True}]

    """

    def __init__(self):
        super().__init__(self._one_hot)

    def _one_hot(self, x):
        for i, xi in x.items():
            if isinstance(xi, str):
                x[f'{i}_{x.pop(i)}'] = True
        return x



import abc

try:
    import surprise
    SURPRISE_INSTALLED = True
except ImportError:
    SURPRISE_INSTALLED = False


if SURPRISE_INSTALLED:

    class Recommender(abc.ABC, surprise.AlgoBase):

        @abc.abstractmethod
        def fit_one(self, row_id, col_id, value):
            pass

        @abc.abstractmethod
        def predict_one(self, row_id, col_id):
            pass

        def fit(self, trainset):
            """Fits to an entire training set.

            This method is compatible with the API from the `surprise` library.

            """
            surprise.AlgoBase.fit(self, trainset)
            for triplet in trainset.all_ratings():
                self.fit_one(*triplet)
            return self

        def estimate(self, *_):
            return self.predict_one(*_)

else:

    class Recommender(abc.ABC):

        @abc.abstractmethod
        def fit_one(self, row_id, col_id, value):
            pass

        @abc.abstractmethod
        def predict_one(self, row_id, col_id):
            pass



from sklearn import utils

from .. import stats

from . import base


__all__ = ['RandomNormal']


class RandomNormal(base.Recommender):
    """

    This is equivalent to `surprise.NormalPredictor`.

    """

    def __init__(self, random_state=42):
        super().__init__()
        self.variance = stats.Variance()
        self.rng = utils.check_random_state(random_state)

    def fit_one(self, r_id, c_id, y):
        y_pred = self.predict_one(r_id, c_id)
        self.variance.update(y)
        return y_pred

    def predict_one(self, r_id, c_id):
        μ = self.variance.mean.get()
        σ = self.variance.get() ** 0.5
        return self.rng.normal(μ, σ)



import collections
import copy

from .. import base
from .. import stats


__all__ = ['GroupBy']


class GroupBy(base.Transformer):
    """

    Example
    -------

        >>> import creme

        >>> X = [
        ...     {'place': 'Taco Bell', 'revenue': 42},
        ...     {'place': 'Burger King', 'revenue': 16},
        ...     {'place': 'Burger King', 'revenue': 24},
        ...     {'place': 'Taco Bell', 'revenue': 58},
        ...     {'place': 'Burger King', 'revenue': 20},
        ...     {'place': 'Taco Bell', 'revenue': 50}
        ... ]

        >>> agg = creme.feature_extraction.GroupBy(
        ...     on='revenue',
        ...     by='place',
        ...     how=creme.stats.Mean()
        ... )

        >>> for x in X:
        ...     print(agg.fit_one(x))
        {'revenue_mean_by_place': 42.0}
        {'revenue_mean_by_place': 16.0}
        {'revenue_mean_by_place': 20.0}
        {'revenue_mean_by_place': 50.0}
        {'revenue_mean_by_place': 20.0}
        {'revenue_mean_by_place': 50.0}

    """

    def __init__(self, on: str, by: str, how: stats.RunningStatistic):
        self.on = on
        self.by = by
        self.how = how
        self.stats = collections.defaultdict(lambda: copy.deepcopy(how))

    def fit_one(self, x, y=None):
        self.stats[x[self.by]].update(x[self.on])
        return self.transform_one(x)

    def transform_one(self, x):
        return {f'{self.on}_{self.how.name}_by_{self.by}': self.stats[x[self.by]].get()}



import collections

from .. import losses
from .. import stats
from .. import optim

from . import base


__all__ = ['SGDBaseline']


class SGDBaseline(base.Recommender):
    """

    This is somewhat equivalent to `surprise.BaselineOnly` with `method='sgd'`. The difference is
    that `surprise` always uses plain SGD whereas with `creme` you can use any SGD method, for
    example Adam.

    """

    def __init__(self, row_optimizer=optim.VanillaSGD(), col_optimizer=optim.VanillaSGD(),
                 loss=losses.SquaredLoss(), l2=0):
        super().__init__()
        self.row_optimizer = row_optimizer
        self.col_optimizer = col_optimizer
        self.loss = loss
        self.l2 = l2
        self.global_mean = stats.Mean()
        self.row_biases = collections.defaultdict(lambda: 0.)
        self.col_biases = collections.defaultdict(lambda: 0.)

    def fit_one(self, r_id, c_id, y):

        # Predict the value
        y_pred = self.predict_one(r_id, c_id)

        # Compute the gradient of the loss
        loss_gradient = self.loss.gradient(y, y_pred)

        # Update the row biases
        r_grad = {r_id: loss_gradient + self.l2 * self.row_biases.get(r_id, 0)}
        self.row_biases = self.row_optimizer.update_weights_with_gradient(self.row_biases, r_grad)

        # Update the row biases
        c_grad = {c_id: loss_gradient + self.l2 * self.col_biases.get(c_id, 0)}
        self.col_biases = self.col_optimizer.update_weights_with_gradient(self.col_biases, c_grad)

        # Update the global mean
        self.global_mean.update(y)

        return y_pred

    def predict_one(self, r_id, c_id):
        return self.global_mean.get() + self.row_biases.get(r_id, 0) + self.col_biases.get(c_id, 0)



import collections

from .. import base
from .. import losses
from .. import optim

from . import util


__all__ = ['LinearRegression']


class LinearRegression(base.Regressor):
    """
    Examples
    --------

        #!python
        >>> import creme.linear_model
        >>> import creme.pipeline
        >>> import creme.preprocessing
        >>> import creme.stream
        >>> from sklearn import datasets
        >>> from sklearn import metrics

        >>> X_y = creme.stream.iter_sklearn_dataset(
        ...     load_dataset=datasets.load_diabetes,
        ...     shuffle=True,
        ...     random_state=42
        ... )
        >>> model = model = creme.pipeline.Pipeline([
        ...     ('scale', creme.preprocessing.StandardScaler()),
        ...     ('bias', creme.preprocessing.BiasAppender()),
        ...     ('learn', creme.linear_model.LinearRegression())
        ... ])
        >>> metric = metrics.mean_squared_error

        >>> #creme.model_selection.online_score(X_y, model, metric)

    """

    def __init__(self, optimizer=optim.VanillaSGD(0.01), loss=losses.SquaredLoss()):
        self.optimizer = optimizer
        self.loss = loss
        self.weights = collections.defaultdict(lambda: 0.)

    def fit_one(self, x, y):

        # Predict the output of the given features
        y_pred = self.predict_one(x)

        # Compute the error gradient
        loss_gradient = self.loss.gradient(y, y_pred)
        gradient = {i: xi * loss_gradient for i, xi in x.items()}

        # Update the weights with the error gradient
        self.weights = self.optimizer.update_weights(self.weights, gradient)

        return y_pred

    def predict_one(self, x):
        return util.dot(x, self.weights)



import collections
import math

from . import base


class MultinomialNB(base.MultiClassifier):
    """

    The input vector has to contain positive values, such as counts or TF-IDF values.

    https://nlp.stanford.edu/IR-book/html/htmledition/naive-bayes-text-classification-1.html

    Example
    -------

    >>> import math
    >>> import creme.feature_extraction
    >>> import creme.naive_bayes
    >>> import creme.pipeline

    >>> docs = [
    ...     ('Chinese Beijing Chinese', 'yes'),
    ...     ('Chinese Chinese Shanghai', 'yes'),
    ...     ('Chinese Macao', 'yes'),
    ...     ('Tokyo Japan Chinese', 'no')
    ... ]
    >>> model = creme.pipeline.Pipeline([
    ...     ('tokenize', creme.feature_extraction.CountVectorizer(on='text', lowercase=False)),
    ...     ('nb', creme.naive_bayes.MultinomialNB(alpha=1))
    ... ])
    >>> for x, y in docs:
    ...     y_pred = model.fit_one({'text': x}, y)

    >>> model.steps[-1][1].p_class('yes')
    0.75
    >>> cp = model.steps[-1][1].p_term_given_class
    >>> cp('Chinese', 'yes') ==  3 / 7
    True
    >>> cp('Tokyo', 'yes') ==  1 / 14
    True
    >>> cp('Japan', 'yes') ==  1 / 14
    True
    >>> cp('Chinese', 'no') ==  2 / 9
    True
    >>> cp('Tokyo', 'no') ==  2 / 9
    True
    >>> cp('Japan', 'no') ==  2 / 9
    True

    >>> new_text = 'Chinese Chinese Chinese Tokyo Japan'
    >>> tokens = model.steps[0][1].transform_one({'text': new_text})
    >>> llh = model.steps[-1][1].calc_log_likelihoods(tokens)
    >>> math.exp(llh['yes'])
    0.0003...
    >>> math.exp(llh['no'])
    0.0001...
    >>> model.predict_one({'text': new_text})
    'yes'

    """

    def __init__(self, alpha=1.0):
        self.alpha = alpha
        self.n = 0
        self.class_counts = collections.defaultdict(lambda: 0)
        self.term_counts = collections.defaultdict(lambda: collections.defaultdict(lambda: 0))
        self.class_term_counts = collections.defaultdict(lambda: 0)

    @property
    def n_terms(self):
        return len(self.term_counts)

    def p_class(self, c):
        """Return P(class)."""
        return self.class_counts.get(c, 0) / self.n

    def p_term_given_class(self, term, c):
        """Return P(term | class)."""
        numerator = self.term_counts[term][c] + self.alpha
        denominator = self.class_term_counts[c] + self.alpha * self.n_terms
        return numerator / denominator

    def fit_one(self, x, y):
        y_pred = self.predict_proba_one(x)
        self.n += 1
        self.class_counts[y] += 1
        for term, frequency in x.items():
            self.term_counts[term][y] += frequency
            self.class_term_counts[y] += frequency
        return y_pred

    def calc_log_likelihoods(self, x):
        return {
            c: math.log(self.p_class(c)) + sum(
                frequency * math.log(self.p_term_given_class(term, c))
                for term, frequency in x.items()
            )
            for c in self.class_counts
        }

    def predict_proba_one(self, x):
        llh = self.calc_log_likelihoods(x)
        total = sum(llh.values())
        return {c: likelihood / total for c, likelihood in llh.items()}

    def predict_one(self, x):
        llh = self.calc_log_likelihoods(x)
        return max(llh, key=llh.get)



from .. import base


__all__ = ['FunctionTransformer']


class FunctionTransformer(base.Transformer):

    def __init__(self, func):
        self.func = func

    def fit_one(self, x, y=None):
        return self.func(x)

    def transform_one(self, x):
        return self.func(x)



import collections

import numpy as np
from sklearn import utils

from .. import base


__all__ = ['KMeans']


def euclidean_distance(a, b):
    return sum((a.get(k, 0) - b.get(k, 0)) ** 2 for k in set([*a.keys(), *b.keys()]))


class KMeans(base.Clusterer):
    """Incremental k-means.

    The most common way to implement batch k-means is to use Lloyd's algorithm, which consists in
    assigning all the data points to a set of cluster centers and then moving the centers
    accordingly. This requires multiple passes over the data and thus isn't applicable in a
    streaming setting.

    In this implementation we start by finding the cluster that is closest to the current
    observation. We then move the cluster's central position towards the new observation. The
    `halflife` parameter determines by how much the movement should be.

    Example
    -------
    In the following example the cluster assignments are exactly the same as when using `sklearn`'s
    batch implementation.

        >>> import creme.cluster
        >>> import numpy as np
        >>> X = np.array([[1, 2], [1, 4], [1, 0],
        ...               [4, 2], [4, 4], [4, 0]])
        >>> k_means = creme.cluster.KMeans(n_clusters=2, halflife=0.4, sigma=3, random_state=42)
        >>> k_means = k_means.fit(X)

        >>> k_means.predict(X)
        array([0, 0, 0, 1, 1, 1], dtype=int32)

        >>> k_means.predict([[0, 0], [4, 4]])
        array([0, 1], dtype=int32)

        >>> k_means.cluster_centers_
        array([[1.17645129, 1.93067455],
        ...    [3.4742213 , 1.84015401]])

    References
    -------
    - [Sequential k-Means Clustering](http://www.cs.princeton.edu/courses/archive/fall08/cos436/Duda/C/sk_means.htm)

    """

    def __init__(self, n_clusters=8, halflife=0.5, mu=0, sigma=1, distance=euclidean_distance,
                 random_state=None):
        self.n_clusters = n_clusters
        """Maximum number of clusters to assign."""
        self.halflife = halflife
        """Amount by which to move the cluster centers, a reasonable value if between 0 and 1."""
        self.mu = mu
        """Mean of the normal distribution used to instantiate cluster positions."""
        self.sigma = sigma
        """Standard deviation of the normal distribution used to instantiate cluster positions."""
        self.distance = distance
        """Metric used to compute distances between an observation and a cluster."""
        self.rng = utils.check_random_state(random_state)
        self.centers = {
            i: collections.defaultdict(lambda: self.rng.normal(self.mu, self.sigma))
            for i in range(n_clusters)
        }

    @property
    def cluster_centers_(self):
        """Returns the cluster centers in the same format as scikit-learn."""
        return np.array([
            list(coords.values())
            for coords in self.centers.values()
        ])

    def fit_one(self, x, y=None):
        # Find the cluster with the closest center
        closest = self.predict_one(x)

        # Move the cluster's center
        for i, xi in x.items():
            self.centers[closest][i] += self.halflife * (xi - self.centers[closest][i])

        return closest

    def predict_one(self, x):
        return min(self.centers, key=lambda c: self.distance(x, self.centers[c]))



import abc
import typing

from . import lr_schedule


class Optimizer(abc.ABC):

    def __init__(self, lr: typing.Union[lr_schedule.LRScheduler, float]):
        self.lr = lr_schedule.ConstantLR(lr) if isinstance(lr, float) else lr
        self.t = 0

    @property
    def learning_rate(self) -> float:
        return self.lr.get(self.t)

    def update_weights_with_gradient(self, w: dict, g: dict) -> dict:
        raise NotImplementedError

    def update_weights(self, x, y, w, f_pred, f_grad):
        """

        This method gives more control over the gradient computation than
        `update_weights_with_gradient`. This is required for optimisation methods that have to
        take a step before actually computing the gradient, such as FTRL-Proximal and Nesterov
        Momentum. However most optimizers could (and should) directly use
        `update_weights_with_gradient`.

        """

        # Predict the output of the given features
        y_pred = f_pred(x, w)

        # Compute the gradient
        gradient = f_grad(y, y_pred, x, w)

        # Update the weights
        w = self.update_weights_with_gradient(w, gradient)

        self.t += 1

        return w, y_pred



from . import base
from . import pipeline


__all__ = ['iter_online_score', 'online_score']


def _identity(y_pred):
    return y_pred


def _get_probas_as_list(y_pred):
    return list(y_pred.values())


def _get_class(y_pred):
    return max(y_pred, key=y_pred.get)


def online_score(X_y, model, metric, use_proba=False):
    """Computes the online score of a model given a stream of data and a metric.

    """

    get_pred = _identity

    if isinstance(model, pipeline.Pipeline):

        if isinstance(model._final_estimator, base.MultiClassifier):
            get_pred = _get_probas_as_list if use_proba else _get_class

    else:

        if isinstance(model, base.MultiClassifier):
            get_pred = _get_probas_as_list if use_proba else _get_class

    y_true, y_pred = [], []

    for x, y in X_y:
        y_true.append(y)
        y_pred.append(get_pred(model.fit_one(x, y)))

    return metric(y_true, y_pred)


def iter_online_score(stream, model, scoring, k=1):

    y_true, y_pred = [], []

    for i, (x, y) in enumerate(stream):
        y_true.append(y)
        y_pred.append(model.fit_one(x, y))
        if i % k == 0:
            yield scoring(y_true, y_pred)



import collections

from . import base


__all__ = ['Adam']


class Adam(base.Optimizer):
    """
    Example
    -------

        #!python
        >>> import creme
        >>> from sklearn import datasets
        >>> from sklearn import metrics

        >>> X_y = creme.stream.iter_sklearn_dataset(
        ...     load_dataset=datasets.load_breast_cancer,
        ...     shuffle=True,
        ...     random_state=42
        ... )
        >>> optimiser = creme.optim.Adam()
        >>> model = creme.pipeline.Pipeline([
        ...     ('scale', creme.preprocessing.StandardScaler()),
        ...     ('learn', creme.linear_model.LogisticRegression(optimiser))
        ... ])
        >>> metric = metrics.roc_auc_score

        >>> creme.model_selection.online_score(X_y, model, metric)
        0.984217...

    """

    def __init__(self, lr=0.1, beta_1=0.9, beta_2=0.999, eps=1e-8):
        super().__init__(lr)
        self.beta_1 = beta_1
        self.beta_2 = beta_2
        self.eps = eps
        self.m = collections.defaultdict(lambda: 0.)
        self.v = collections.defaultdict(lambda: 0.)

    def update_weights_with_gradient(self, w, g):

        for i, gi in g.items():
            self.m[i] = self.beta_1 * self.m[i] + (1 - self.beta_1) * gi
            self.v[i] = self.beta_2 * self.v[i] + (1 - self.beta_2) * gi ** 2
            m = self.m[i] / (1 - self.beta_1 ** (self.t + 1))
            v = self.v[i] / (1 - self.beta_2 ** (self.t + 1))
            w[i] -= self.learning_rate * m / (v ** 0.5 + self.eps)

        return w



import collections

from . import base


__all__ = ['NesterovMomentum']


class NesterovMomentum(base.Optimizer):
    """
    Example
    -------

        #!python
        >>> import creme
        >>> from sklearn import datasets
        >>> from sklearn import metrics

        >>> X_y = creme.stream.iter_sklearn_dataset(
        ...     load_dataset=datasets.load_breast_cancer,
        ...     shuffle=True,
        ...     random_state=42
        ... )
        >>> optimiser = creme.optim.NesterovMomentum()
        >>> model = creme.pipeline.Pipeline([
        ...     ('scale', creme.preprocessing.StandardScaler()),
        ...     ('learn', creme.linear_model.LogisticRegression(optimiser))
        ... ])
        >>> metric = metrics.roc_auc_score

        >>> creme.model_selection.online_score(X_y, model, metric)
        0.975166...

    """

    def __init__(self, lr=0.1, rho=0.9):
        super().__init__(lr)
        self.rho = rho
        self.s = collections.defaultdict(lambda: 0.)

    def update_weights(self, x, y, w, f_pred, f_grad):

        # Move the weights to the future position
        for i in w:
            w[i] -= self.rho * self.s[i]

        # Compute the gradient
        y_pred = f_pred(x, w)
        gradient = f_grad(y, y_pred, x, w)

        # Update the step and the weights
        for i, gi in gradient.items():
            self.s[i] = self.rho * self.s[i] + self.learning_rate * gi
            w[i] -= self.s[i]

        return w, y_pred



from .constant import BiasAppender
from .function import FunctionTransformer
from .hash import FeatureHasher
from .one_hot import OneHotEncoder
from .scale import StandardScaler


__all__ = [
    'BiasAppender',
    'FeatureHasher',
    'FunctionTransformer',
    'OneHotEncoder',
    'StandardScaler'
]



import collections

from sklearn import utils

from .. import base


__all__ = ['FeatureHasher']


class FeatureHasher(base.Transformer):
    """
    Example
    -------

        #!python
        >>> import creme
        >>> hasher = creme.preprocessing.FeatureHasher(n_features=10, random_state=0)
        >>> X = [{'dog': 1, 'cat':2, 'elephant':4},{'dog': 2, 'run': 5}]
        >>> for x in X:
        ...     print(hasher.transform_one(x))
        defaultdict(<class 'int'>, {2: -1, 1: 2, 8: 4})
        defaultdict(<class 'int'>, {2: -2, 4: 5})

    """

    def __init__(self, n_features=1048576, alternate_sign=True, encoding='utf-8',
                 random_state=None):
        self.n_features = n_features
        self.alternate_sign = alternate_sign
        self.encoding = encoding
        self.rng = utils.check_random_state(random_state)
        self.seed = self.rng.randint(0, 2 ** 32 - 1)

    def fit_one(self, x, y=None):
        return self.transform_one(x)

    def hash_str(self, s):
        return utils.murmurhash3_32(bytes(s, self.encoding), self.seed)

    def transform_one(self, x):

        x_hashed = collections.defaultdict(int)

        for feature, value in x.items():

            if isinstance(value, str):
                feature = f'{feature}={value}'
                value = 1

            if value == 0:
                continue

            h = self.hash_str(feature)
            i = h % self.n_features

            if self.alternate_sign:
                value *= (h >= 0) * 2 - 1

            x_hashed[i] += value

        return x_hashed



"""
To be clear, online gradient descent and stochastic gradient descent are the same thing.

- http://ruder.io/optimizing-gradient-descent/
- https://keras.io/optimisers/
"""
from .ada_delta import AdaDelta
from .ada_grad import AdaGrad
from .adam import Adam
from .ftrl import FTRLProximal
from .lr_schedule import ConstantLR
from .lr_schedule import LinearDecreaseLR
from .momentum import Momentum
from .nesterov import NesterovMomentum
from .rms_prop import RMSProp
from .vanilla_sgd import VanillaSGD


__all__ = [
    'AdaDelta',
    'AdaGrad',
    'Adam',
    'ConstantLR',
    'FTRLProximal',
    'LinearDecreaseLR',
    'Momentum',
    'NesterovMomentum',
    'RMSProp',
    'VanillaSGD'
]



import collections

from .. import base
from .. import losses
from .. import optim

from . import util


__all__ = ['LogisticRegression']


class LogisticRegression(base.BinaryClassifier):

    def __init__(self, optimizer=optim.VanillaSGD(), loss=losses.LogLoss(), l2=0):
        self.optimizer = optimizer
        self.loss = loss
        self.l2 = l2
        self.weights = collections.defaultdict(lambda: 0.)

    def _predict_proba_one_with_weights(self, x, w):
        return util.sigmoid(util.dot(x, w))

    def _calc_gradient(self, y_true, y_pred, x, w):
        loss_gradient = self.loss.gradient(y_true, y_pred)
        return {i: xi * loss_gradient + self.l2 * w.get(i, 0) for i, xi in x.items()}

    def fit_one(self, x, y):

        # Update the weights with the error gradient
        self.weights, y_pred = self.optimizer.update_weights(
            x=x,
            y=y,
            w=self.weights,
            f_pred=self._predict_proba_one_with_weights,
            f_grad=self._calc_gradient
        )

        return y_pred

    def predict_proba_one(self, x):
        return self._predict_proba_one_with_weights(x, self.weights)

    def predict_one(self, x):
        return self.predict_proba(x) > 0.5



import collections

import numpy as np

from . import base


__all__ = ['FTRLProximal']


class FTRLProximal(base.Optimizer):
    """
    Example
    -------

        #!python
        >>> import creme
        >>> from sklearn import datasets
        >>> from sklearn import metrics

        >>> X_y = creme.stream.iter_sklearn_dataset(
        ...     load_dataset=datasets.load_breast_cancer,
        ...     shuffle=True,
        ...     random_state=42
        ... )
        >>> optimiser = creme.optim.FTRLProximal()
        >>> model = creme.pipeline.Pipeline([
        ...     ('scale', creme.preprocessing.StandardScaler()),
        ...     ('learn', creme.linear_model.LogisticRegression(optimiser))
        ... ])
        >>> metric = metrics.roc_auc_score

        >>> creme.model_selection.online_score(X_y, model, metric)
        0.989859...

    """

    def __init__(self, alpha=.05, beta=1., l1=0., l2=1.):
        self.alpha = alpha
        self.beta = beta
        self.l1 = l1
        self.l2 = l2
        self.z = collections.defaultdict(lambda: 0.)
        self.n = collections.defaultdict(lambda: 0.)

    def update_weights(self, x, y, w, f_pred, f_grad):

        alpha = self.alpha
        beta = self.beta
        l1 = self.l1
        l2 = self.l2
        z = self.z
        n = self.n
        w = {}

        for i, xi in x.items():
            if abs(z[i]) > l1:
                w[i] = -((beta + n[i] ** 0.5) / alpha + l2) ** -1 * (z[i] - np.sign(z[i]) * l1)

        y_pred = f_pred(x, w)
        gradient = f_grad(y, y_pred, x, w)

        for i, gi in gradient.items():
            s = ((n[i] + gi ** 2) ** 0.5 - n[i] ** 0.5) / alpha
            z[i] += gi - s * w.get(i, 0)
            n[i] += gi ** 2

        return w, y_pred



import collections

from sklearn import pipeline
from sklearn.utils import metaestimators

from . import base


__all__ = ['Pipeline']


class Pipeline(pipeline.Pipeline):

    @property
    def _final_estimator(self):
        return self.steps[-1][1]

    def fit_one(self, x, y):
        for _, step in self.steps:
            x = step.fit_one(x, y)
        return x

    @metaestimators.if_delegate_has_method(delegate='_final_estimator')
    def transform_one(self, x):
        for _, step in self.steps:
            x = step.transform_one(x)
        return x

    @metaestimators.if_delegate_has_method(delegate='_final_estimator')
    def predict_one(self, x):
        for _, step in self.steps[:-1]:
            x = step.transform_one(x)
        return self.steps[-1][1].predict_one(x)

    @metaestimators.if_delegate_has_method(delegate='_final_estimator')
    def predict_proba_one(self, x):
        for _, step in self.steps[:-1]:
            x = step.transform_one(x)
        return self.steps[-1][1].predict_proba_one(x)


class FeatureUnion(base.Transformer):
    """
    Example
    -------

        >>> import creme.feature_extraction
        >>> import creme.pipeline

        >>> X = [
        ...     {'place': 'Taco Bell', 'revenue': 42},
        ...     {'place': 'Burger King', 'revenue': 16},
        ...     {'place': 'Burger King', 'revenue': 24},
        ...     {'place': 'Taco Bell', 'revenue': 58},
        ...     {'place': 'Burger King', 'revenue': 20},
        ...     {'place': 'Taco Bell', 'revenue': 50}
        ... ]

        >>> mean = creme.feature_extraction.GroupBy(
        ...     on='revenue',
        ...     by='place',
        ...     how=creme.stats.Mean()
        ... )
        >>> count = creme.feature_extraction.GroupBy(
        ...     on='revenue',
        ...     by='place',
        ...     how=creme.stats.Count()
        ... )
        >>> agg = creme.pipeline.FeatureUnion([mean, count])

        >>> for x in X:
        ...     print(sorted(agg.fit_one(x).items()))
        [('revenue_count_by_place', 1), ('revenue_mean_by_place', 42.0)]
        [('revenue_count_by_place', 1), ('revenue_mean_by_place', 16.0)]
        [('revenue_count_by_place', 2), ('revenue_mean_by_place', 20.0)]
        [('revenue_count_by_place', 2), ('revenue_mean_by_place', 50.0)]
        [('revenue_count_by_place', 3), ('revenue_mean_by_place', 20.0)]
        [('revenue_count_by_place', 3), ('revenue_mean_by_place', 50.0)]

    """

    def __init__(self, transformers):
        self.transformers = transformers

    def fit_one(self, x, y=None):
        return dict(collections.ChainMap(*(t.fit_one(x, y) for t in self.transformers)))

    def transform_one(self, x):
        return dict(collections.ChainMap(*(t.transform_one(x) for t in self.transformers)))



from . import function


__all__ = ['BiasAppender']


class BiasAppender(function.FunctionTransformer):
    """
    Example
    -------

        #!python
        >>> import creme

        >>> x = {'x': 42}
        >>> creme.preprocessing.BiasAppender().fit_one(x)
        {'x': 42, 'bias': 1.0}

    """

    def __init__(self, name='bias'):
        super().__init__(self._add_bias)
        self.name = name

    def _add_bias(self, x):
        x[self.name] = 1.
        return x



"""

- [Decision Trees and Forests: A Probabilistic Perspective](http://www.gatsby.ucl.ac.uk/~balaji/balaji-phd-thesis.pdf)
- [Mondrian Forests for Large-Scale Regression when Uncertainty Matters](http://www.gatsby.ucl.ac.uk/~balaji/mfregression_aistats16.pdf)
- [Mondrian Forests: Efficient Online Random Forests](https://papers.nips.cc/paper/5234-mondrian-forests-efficient-online-random-forests.pdf)
- [Intuition behind Mondrian Trees](https://scikit-garden.github.io/examples/MondrianTreeRegressor/)
- [Mondrian Forest](https://ldocao.wordpress.com/2016/08/26/mondrian-forest/)
"""
import collections
import copy
import math
import random

import numpy as np
from sklearn import utils


Split = collections.namedtuple('Split', 'feature threshold')

Bound = collections.namedtuple('Bound', 'lower upper')


class Block(collections.defaultdict):

    def __init__(self, bounds=None):
        super().__init__(lambda: Bound(0., 0.))
        for feature, bound in (bounds or {}).items():
            self[feature] = bound

    def contains(self, x):
        """
        Examples
        --------
        >>> block = Block({'x': Bound(-1, 1)})
        >>> block.contains({'x': 0.5})
        True
        >>> block.contains({'x': -2})
        False
        >>> block.contains({'x': 2})
        False
        """
        if self:
            return all(bound.lower <= x[feature] <= bound.upper for feature, bound in self.items())
        return True


class MondrianNode:

    def __init__(self, tau, feature, threshold, left=None, right=None):
        self.tau = tau
        self.feature = feature
        self.threshold = threshold
        self.left = left
        self.right = right

    @property
    def is_leaf(self):
        return self.left is None and self.right is None

    def extend(self, x):
        e_l = {i: max(self.block[i].lower - xi, 0) for i, xi in x.items()}
        e_u = {i: max(xi - self.block[i].upper, 0) for i, xi in x.items()}

        #
        rate = sum(e_l.values()) + sum(e_u.values())
        E = np.random.exponential(scale=1 / rate, size=1)[0]

        self.block = new_block
        return self

    def predict(self, x):

        if x[self.feature] < self.threshold:
            return self.left.predict(x)
        return self.right.predict(x)


        #rate = sum(u - l for u, l in zip(self.lower_bounds, self.upper_bounds))
        #    time = np.random.exponential(scale=1 / rate, size=1)[0]


class MondrianTree:

    def __init__(self, lifetime, random_state=None):
        self.lifetime = lifetime
        self.lower_bounds = collections.defaultdict(float)
        self.upper_bounds = collections.defaultdict(float)
        self.rng = utils.check_random_state(random_state)

    def _sample_mondrian_block(self, x):
        rate = 2 * sum(x.values())
        E = np.random.exponential(scale=1 / rate, size=1)[0]
        if E < self.lifetime:
            feature = random.choice(features)
            return MondrianNode(tau=E, feature=feature, threshold=x[feature])
        return MondrianNode(tau=self.lifetime, feature=None, threshold=None)

    #def fit(self, x, y):
    #    self.tree_ = self._sample_mondrian_block()




import math


__all__ = ['dot', 'sigmoid']


def dot(x: dict, y: dict):
    """Computes the dot product of two vectors represented as dicts.

    Examples
    --------
    >>> x = {'x0': 1, 'x1': 2}
    >>> y = {'x1': 21, 'x2': 3}
    >>> dot(x, y)
    42
    """
    return sum(xi * y.get(i, 0) for i, xi in min(x, y, key=len).items())


def sigmoid(x: float):
    return 1 / (1 + math.exp(-x))
